#include <vcl.h>
#pragma hdrstop
#include "E7_DDraw.h"
#include "errors7.h"
#pragma package(smart_init)

// ValidCtrCheck is used to assure that the components created do not have
// any pure virtual functions.
//
static inline void ValidCtrCheck(TE7_DDraw *)
{
  new TE7_DDraw(NULL);
}
//---------------------------------------------------------------------------
TE7_DDraw *AE_DDraw;
//---------------------------------------------------------------------------
void Timer2Timer(HWND H, UINT Msg, UINT Event, DWORD Time)
{
  if (AE_DDraw->Active) AE_DDraw->Flip();
}
//---------------------------------------------------------------------------
namespace DDUtils
{
void __fastcall DDBlitFast(LPDIRECTDRAWSURFACE7 Dest, LPDIRECTDRAWSURFACE7 Source,
 IPoint SourcePos, RECT SourceSize, LPRECT DestSize)
{
  HRESULT ddrval;

  if (DestSize == NULL) DestSize = &AE_DDraw->ScreenRect;
  if (SourcePos.X < DestSize->left) return;
  else if (SourcePos.Y < DestSize->top) return;
  else if (SourcePos.X > DestSize->right) return;
  else if (SourcePos.Y > DestSize->bottom) return;

                                                 //Ready to draw the image
  ddrval = Dest->BltFast(SourcePos.X, SourcePos.Y, Source, &SourceSize,
   DDBLTFAST_SRCCOLORKEY);
  if (ddrval == DDERR_INVALIDRECT) return;       //Invalid rectangle ok
  if (ddrval != DD_OK) AE_DDraw->DDTest(ddrval, "BltFast");
}
//---------------------------------------------------------------------------
void __fastcall DDBlitFx(LPDIRECTDRAWSURFACE7 Dest, LPDIRECTDRAWSURFACE7 Source,
 IPoint SourcePos, RECT Size, DWORD Blt_Flags, LPDDBLTFX ddfx)
{
  /*******************************************************/
  /*This method is recomended for advanced users         */
  /*For a transparent color use DDBLT_KEYSRC in Blt_Flags*/
  /*To use DDBLTFX use DDBLT_DDFX in Blt_Flags           */
  /*For a list of all the DDBLTFX flags see DDraw.h      */
  /*******************************************************/
  HRESULT ddrval;
  RECT Pos;

  Pos.left = SourcePos.X + Size.left;            //Setup Pos
  Pos.right = Pos.left + Size.right;
  Pos.top = SourcePos.Y + Size.top;
  Pos.bottom = Pos.top + Size.bottom;

                                                 //Ready to draw the image
  ddrval = Dest->Blt(&Pos, Source, &Size, Blt_Flags, ddfx);
  if (ddrval == DDERR_INVALIDRECT) return;       //Invalid rectangle ok
  if (ddrval != DD_OK) AE_DDraw->DDTest(ddrval, "BltFx");
}
//---------------------------------------------------------------------------
HRESULT DDCopyBitmap(LPDIRECTDRAWSURFACE7 lpdds, HBITMAP hbm, RECT Grid)
{
  HDC hdcImage;
  HDC hdc;
  BITMAP bm;
  DDSURFACEDESC2 ddsd;
  HRESULT hr;

  if (hbm == NULL) throw EDDError("DDCopyBitmap: No Bitmap available");
  if (lpdds == NULL) throw EDDError("DDCopyBitmap: No Surface available");

  lpdds->Restore();
  hdcImage = CreateCompatibleDC(NULL);
  if (!hdcImage) throw EDDError("Create Compatible DC failed!");

  SelectObject(hdcImage, hbm);

  GetObject(hbm, sizeof(bm), &bm);               //get size of bitmap
  if (Grid.right == 0) Grid.right = bm.bmWidth;  //use the passed size, unless zero
  if (Grid.bottom == 0) Grid.bottom = bm.bmHeight;
  ddsd.dwSize = sizeof(ddsd);
  ddsd.dwFlags = DDSD_HEIGHT | DDSD_WIDTH;
  lpdds->GetSurfaceDesc(&ddsd);

  if ((hr = lpdds->GetDC(&hdc)) == DD_OK) {
    StretchBlt(hdc, 0, 0, ddsd.dwWidth, ddsd.dwHeight, hdcImage, Grid.left,
     Grid.top, Grid.right, Grid.bottom, SRCCOPY);
    lpdds->ReleaseDC(hdc);
  }
  DeleteDC(hdcImage);

  return hr;
}
//---------------------------------------------------------------------------
LPDIRECTDRAWSURFACE7 DDCreateSurface(DWORD Width, DWORD Height)
{
  DDSURFACEDESC2 ddsd;
  LPDIRECTDRAWSURFACE7 lpdds;
  HRESULT ddrval;

  ZeroMemory(&ddsd, sizeof(ddsd));
  ddsd.dwSize = sizeof(ddsd);                    //Setup Surface
  ddsd.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
  ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
  ddsd.dwHeight = Height;
  ddsd.dwWidth = Width;
  ddrval = AE_DDraw->ADirectDraw->CreateSurface(&ddsd, &lpdds, NULL);
  if (ddrval != DD_OK) AE_DDraw->DDTest(ddrval,
   "DDCreateSurface: Create surface Failed");

  return lpdds;
}
//---------------------------------------------------------------------------
/*LPDIRECTDRAWCLIPPER CreateClipper(LPDIRECTDRAW7 pdd, HWND Handle)
{
  HRESULT hr;
  LPDIRECTDRAWCLIPPER lpClipper;

  hr = pdd->CreateClipper(0, &lpClipper, NULL);
  if (hr != DD_OK) throw EDDError("CreateClipper: No Clipper");
  hr = lpClipper->SetHWnd(0, Handle);

  if (hr != DD_OK) throw EDDError("CreateClipper: Can't set clipper window handle");

  return lpClipper;
}*/
//---------------------------------------------------------------------------
DWORD DDColorMatch(LPDIRECTDRAWSURFACE7 lpdds, COLORREF rgb)
{
  COLORREF rgbT;
  HDC hdc;
  DWORD dw = CLR_INVALID;
  DDSURFACEDESC2 ddsd;
  HRESULT hres;

  if (rgb != CLR_INVALID && lpdds->GetDC(&hdc) == DD_OK) {
    rgbT = GetPixel(hdc, 0, 0);                  //save current pixel value
    SetPixel(hdc, 0, 0, rgb);                    //set the value
    lpdds->ReleaseDC(hdc);
  }

  ddsd.dwSize = sizeof(ddsd);
  while ((hres = lpdds->Lock(NULL, &ddsd, 0, NULL)) == DDERR_WASSTILLDRAWING);
  if (hres == DD_OK) {
    dw  = *(DWORD *)ddsd.lpSurface;
    if(ddsd.ddpfPixelFormat.dwRGBBitCount < 32)
    dw &= (1 << ddsd.ddpfPixelFormat.dwRGBBitCount)-1;
    lpdds->Unlock(NULL);
  }
  if (rgb != CLR_INVALID && lpdds->GetDC(&hdc) == DD_OK) {
    SetPixel(hdc, 0, 0, rgbT);
    lpdds->ReleaseDC(hdc);
  }
  return dw;
} //DDColorMatch
//---------------------------------------------------------------------------
void __fastcall DDDestroySurface(LPDIRECTDRAWSURFACE7 &lpdds)
{
  if (lpdds != NULL) {
    lpdds->Release();
    lpdds = NULL;
  }
}
//---------------------------------------------------------------------------
IPoint DDGetImageSize(HANDLE Dll, AnsiString Pic)
{
  HBITMAP Bitmap;
  BITMAP bm;
  IPoint Image_Size;

  if (Dll)
   Bitmap = LoadImage(Dll, Pic.c_str(),
    IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);
  else
   Bitmap = LoadImage(GetModuleHandle(NULL), Pic.c_str(),
    IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);

  if (!Bitmap)
   Bitmap = LoadImage(0, Pic.c_str(), IMAGE_BITMAP,
    0, 0, LR_LOADFROMFILE | LR_CREATEDIBSECTION);

  if (!Bitmap) throw Exception("DDGetImageSize: Unable to load bitmap %s",
   OPENARRAY(TVarRec, (Pic)));

  GetObject(Bitmap, sizeof(bm), &bm);
  Image_Size.X = bm.bmWidth;
  Image_Size.Y = bm.bmHeight;

  return Image_Size;
}
//---------------------------------------------------------------------------
LPDIRECTDRAWSURFACE7 DDLoadBitmap(HANDLE Dll, AnsiString Pic, bool Transparent,
 COLORREF TransColour)
{
  LPDIRECTDRAWSURFACE7 lpdds;
  IPoint ISize;
  int dx, dy;

  ISize = DDGetImageSize(Dll, Pic);              //Get Image Size
  dx = ISize.X;                                  //Width
  dy = ISize.Y;                                  //Height
  lpdds = DDCreateSurface(dx, dy);               //Create the surface

  DDReloadBitmap(lpdds, Dll, Pic);               //Load the picture

  if (Transparent) DDSetColorKey(lpdds, TransColour);

  return lpdds;
}
//---------------------------------------------------------------------------
LPDIRECTDRAWPALETTE DDLoadPalette(HANDLE Instance,
 LPDIRECTDRAW7 pdd, const AnsiString BitmapName)
{
  LPDIRECTDRAWPALETTE ddpal;
  int                 i;
  int                 n;
  int                 fh;
  HRSRC               h;
  LPBITMAPINFO        BitmapInfo;
  PALETTEENTRY        ape[256];
  RGBQUAD *           RGBQuad;

  //build a 332 palette as the default-----------\\
  for (i = 0; i < 256; i++) {
    ape[i].peRed   = (BYTE)(((i >> 5) & 0x07) * 255 / 7);
    ape[i].peGreen = (BYTE)(((i >> 2) & 0x07) * 255 / 7);
    ape[i].peBlue  = (BYTE)(((i >> 0) & 0x03) * 255 / 3);
    ape[i].peFlags = (BYTE)0;
  }

  if (BitmapName == "") throw EDDError("DDLoadPalette: No bitmapname");
  //get a pointer to the bitmap resource---------\\
  if (Instance) {
    h = FindResource(Instance, BitmapName.c_str(), RT_BITMAP);
    if (h) {
      BitmapInfo = (LPBITMAPINFO)LockResource(LoadResource(Instance, h));
      if (!BitmapInfo) throw EDDError("DDLoadPalette: Can't Lock Resouce");
      RGBQuad = (RGBQUAD*)BitmapInfo->bmiColors;

      if (BitmapInfo == NULL || BitmapInfo->bmiHeader.biSize < sizeof(BITMAPINFOHEADER))
       n = 0;
      else if (BitmapInfo->bmiHeader.biBitCount > 8)
       n = 0;
      else if (BitmapInfo->bmiHeader.biClrUsed == 0)
       n = 1 << BitmapInfo->bmiHeader.biBitCount;
      else
       n = BitmapInfo->bmiHeader.biClrUsed;

      //  a DIB color table has its colors stored BGR not RGB
      //  so flip them around.

      for(i = 0; i < n; i++) {
        ape[i].peRed   = RGBQuad[i].rgbRed;
        ape[i].peGreen = RGBQuad[i].rgbGreen;
        ape[i].peBlue  = RGBQuad[i].rgbBlue;
        ape[i].peFlags = 0;
      }
    }
  }
  else {
    fh = _lopen(BitmapName.c_str(), OF_READ);
    if (fh != -1) {
      BITMAPFILEHEADER bf;
      BITMAPINFOHEADER bi;

      _lread(fh, &bf, sizeof(bf));
      _lread(fh, &bi, sizeof(bi));
      _lread(fh, ape, sizeof(ape));
      _lclose(fh);

      if (bi.biSize != sizeof(BITMAPINFOHEADER))
       n = 0;
      else if (bi.biBitCount > 8)
       n = 0;
      else if (bi.biClrUsed == 0)
       n = 1 << bi.biBitCount;
      else
       n = bi.biClrUsed;

      //  a DIB color table has its colors stored BGR not RGB
      //  so flip them around.
      for(i = 0; i < n; i++) {
        BYTE r = ape[i].peRed;
        ape[i].peRed  = ape[i].peBlue;
        ape[i].peBlue = r;
      }
    }
  }

  pdd->CreatePalette(DDPCAPS_8BIT, ape, &ddpal, NULL);

  return ddpal;
}
//---------------------------------------------------------------------------
void __fastcall DDReloadBitmap(LPDIRECTDRAWSURFACE7 pdds, HANDLE Dll,
 AnsiString Pic)
{
  HBITMAP Bitmap;

  if (Dll)
   Bitmap = LoadImage(Dll, Pic.c_str(),
    IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);
  else
   Bitmap = LoadImage(GetModuleHandle(NULL), Pic.c_str(),
    IMAGE_BITMAP, 0, 0, LR_CREATEDIBSECTION);

  if (!Bitmap)
   Bitmap = LoadImage(0, Pic.c_str(), IMAGE_BITMAP,
    0, 0, LR_LOADFROMFILE | LR_CREATEDIBSECTION);

  if (!Bitmap) throw Exception("DDReLoadBitmap: Unable to load bitmap %s",
   OPENARRAY(TVarRec, (Pic)));

  DDCopyBitmap(pdds, Bitmap, Rect(0, 0, 0, 0));
  DeleteObject(Bitmap);
}
//---------------------------------------------------------------------------
HRESULT DDSetColorKey(LPDIRECTDRAWSURFACE7 lpdds, COLORREF Colour)
{
  DDCOLORKEY ddck;
  ddck.dwColorSpaceLowValue  = DDColorMatch(lpdds, Colour);
  ddck.dwColorSpaceHighValue = ddck.dwColorSpaceLowValue;

  return lpdds->SetColorKey(DDCKEY_SRCBLT, &ddck);
}
//---------------------------------------------------------------------------
}; // end namespace

//TE7_DDraw-------------------------------------------------------------------
__fastcall TE7_DDraw::TE7_DDraw(TComponent* Owner)
:TComponent(Owner)
{
  FHandle = ((TWinControl *)(Owner))->Handle;
  FFirstTime = true;
  FDirectDraw = NULL;
  FBitsPerPixel = 16;
}
//---------------------------------------------------------------------------
__fastcall TE7_DDraw::~TE7_DDraw()
{
  DDDestroySurface(FBackSurface);
  DDDestroySurface(FPrimarySurface);
  if(FDirectDraw) FDirectDraw->Release();
}
//---------------------------------------------------------------------------
int TE7_DDraw::ChoseColourDepth(int Depth)
{
  //Can be 8, 16, 32
  if ((Depth != 8) | (Depth != 16) | (Depth != 32)) Depth = 32;
  else if (Depth > 8) Depth /= 2;
  else Depth = 8;

  return Depth;
}
//---------------------------------------------------------------------------
void TE7_DDraw::CreateBack()
{
  DDSCAPS2 DDSCaps;

  if (!FExclusive)
  {
    FBackSurface = DDCreateSurface(ScreenRect.right, ScreenRect.bottom);
    DDTest(FDirectDraw->CreateClipper(0, &FClipper, NULL), "CreateBack");
    FClipper->SetHWnd(0, FHandle);             //Set HWnd to clipper
    DDTest(FPrimarySurface->SetClipper(FClipper), "CreateBack");
  }
  else {
    ZeroMemory(&DDSCaps, sizeof(DDSCaps));
    DDSCaps.dwCaps = DDSCAPS_BACKBUFFER;
    DDTest(FPrimarySurface->GetAttachedSurface(&DDSCaps, &FBackSurface), "CreateBack");
  }
}
//---------------------------------------------------------------------------
void TE7_DDraw::CreateDD()
{
  HRESULT ddrval;
  DWORD Flags;
  bool Done = false;

  CoInitialize(NULL);
  ddrval = DirectDrawCreateEx(NULL, (void**)&FDirectDraw, IID_IDirectDraw7, NULL);

  if(!FAILED(ddrval)) ddrval = FDirectDraw->Initialize(NULL);
  else throw EDDError("Can't Create DirectDraw Object");

  if (!FExclusive) Flags = DDSCL_NORMAL;
  else Flags = DDSCL_EXCLUSIVE | DDSCL_FULLSCREEN | DDSCL_ALLOWREBOOT;

  if (FResolution == trNormal) {                 //Use desktop settings
    DDSURFACEDESC2 ddsd;
    ddsd.dwSize = sizeof(ddsd);
    DDTest(FDirectDraw->GetDisplayMode(&ddsd), "CreateDD");
    BitsPerPixel = ddsd.ddpfPixelFormat.dwRGBBitCount;
    ScreenRect = Rect(0, 0, ddsd.dwWidth, ddsd.dwHeight);
  }

  DDTest(FDirectDraw->SetCooperativeLevel(FHandle, Flags), "CreateDD");

  do {
    if (FResolution != trNormal) SetScreenSize();//Set Resolution
    ddrval = FDirectDraw->SetDisplayMode(ScreenRect.right, ScreenRect.bottom,
     FBitsPerPixel, 0, 0);

    if (ddrval == DD_OK) Done = true;
    else if (ddrval == DDERR_NOEXCLUSIVEMODE) {
      DDTest(FDirectDraw->RestoreDisplayMode(), "CreateDD");
      FBitsPerPixel = ChoseColourDepth(FBitsPerPixel);
    }
    else if (ddrval == DDERR_INVALIDMODE) {
      DDTest(FDirectDraw->RestoreDisplayMode(), "CreateDD");
      if (FResolution > 0) int(FResolution -= 1);
      else ErrorEvent("Can't set display mode");
    }
    else DDTest(ddrval, "CreateDD");

  } while(!Done);

  CreatePrimary();                               //Create Primary Surface
  CreateBack();                                  //Create Back Surface
}
//---------------------------------------------------------------------------
void TE7_DDraw::CreatePrimary()
{
  DDSURFACEDESC2 SurfaceDesc;
  HRESULT ddrval;

  ZeroMemory(&SurfaceDesc, sizeof(SurfaceDesc));
  SurfaceDesc.dwSize = sizeof(SurfaceDesc);

  if (!FExclusive) {
    SurfaceDesc.dwFlags = DDSD_CAPS;
    SurfaceDesc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE;
  }
  else{
    SurfaceDesc.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;
    SurfaceDesc.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE |
                                 DDSCAPS_FLIP |
                                 DDSCAPS_COMPLEX;
    SurfaceDesc.dwBackBufferCount = 1;
  };

  ddrval = FDirectDraw->CreateSurface(&SurfaceDesc, &FPrimarySurface, NULL);
  DDTest(ddrval, "CreatePrimary");
}
//---------------------------------------------------------------------------
void __fastcall TE7_DDraw::EndExclusive()
{
  FActive = false;
  FDirectDraw->FlipToGDISurface();
  if (FExclusive) FDirectDraw->SetCooperativeLevel(FHandle, DDSCL_NORMAL);
}
//---------------------------------------------------------------------------
void __fastcall TE7_DDraw::ErrorEvent(String Msg)
{
  EndExclusive();
  throw EDDError(Msg);
}
//---------------------------------------------------------------------------
void __fastcall TE7_DDraw::Init()
{
  AE_DDraw = this;

  if (FFirstTime) {
    CreateDD();                                  //Create Direct Draw
    FFirstTime = false;
  }

  if ((E_DDSurface) && (E_DDSurface->FBackgroundBitmap != ""))
   E_DDSurface->SetupSurfaces(this);

  if (FUseTimer)
   SetTimer(FHandle, Timer1, FTimerInterval, (FARPROC)Timer2Timer);

  FActive = true;
  Flip();
}
//---------------------------------------------------------------------------
void __fastcall TE7_DDraw::DDTest(HRESULT ddrval, AnsiString Orign)
{
  if (!Windows::Succeeded(ddrval))
  {
    EndExclusive();
    throw EDDError("DDTest Error: %s :%d %s",
      OPENARRAY(TVarRec, (Orign, int(ddrval), GetOleError(ddrval))));
  }
}
//---------------------------------------------------------------------------
void __fastcall TE7_DDraw::SetScene(TE7_DDSurface *E_DDSurface)
{
  if (E_DDSurface) E_DDSurface->DestroyObjects();
  FE_DDSurface = E_DDSurface;
}
//---------------------------------------------------------------------------
void __fastcall TE7_DDraw::Notification(TComponent* AComponent, TOperation Operation)
{
  TComponent::Notification(AComponent, Operation);
  if ((AComponent == FE_DDSurface) && (Operation == opRemove))
   E_DDSurface = NULL;
}
//---------------------------------------------------------------------------
void TE7_DDraw::SetScreenSize()
{
  if (FResolution == tr320x240) ScreenRect = Rect(0, 0, 320, 240);
  else if (FResolution == tr400x300) ScreenRect = Rect(0, 0, 400, 300);
  else if (FResolution == tr512x384) ScreenRect = Rect(0, 0, 512, 384);
  else if (FResolution == tr640x480) ScreenRect = Rect(0, 0, 640, 480);
  else if (FResolution == tr800x600) ScreenRect = Rect(0, 0, 800, 600);
  else if (FResolution == tr960x720) ScreenRect = Rect(0, 0, 960, 720);
  else if (FResolution == tr1024x768) ScreenRect = Rect(0, 0, 1024, 768);
  else if (FResolution == tr1152x864) ScreenRect = Rect(0, 0, 1152, 864);
  else if (FResolution == tr1280x1024) ScreenRect = Rect(0, 0, 1280, 1024);
  else if (FResolution == tr1600x1200) ScreenRect = Rect(0, 0, 1600, 1200);
}
//---------------------------------------------------------------------------
void __fastcall TE7_DDraw::DrawBitmaps()
{
  if (E_DDSurface) E_DDSurface->DrawScene();
}
//---------------------------------------------------------------------------
bool __fastcall TE7_DDraw::MakeItSo(HRESULT DDResult)
{
  bool Result;

  switch(DDResult)
  {
    case DD_OK:
      Result = true;
      break;
    case DDERR_SURFACELOST:
      Result = (RestoreSurfaces() == DD_OK);
      break;
    default:
      Result = DDResult != DDERR_WASSTILLDRAWING;
      break;
  };
  return Result;
}
//---------------------------------------------------------------------------
HRESULT TE7_DDraw::RestoreSurfaces()
{
  HRESULT ddrval;

  ddrval = FPrimarySurface->Restore();
  if ((ddrval == DD_OK) && (E_DDSurface))
   ddrval = E_DDSurface->RestoreSurfaces();
  return ddrval;
}
//---------------------------------------------------------------------------
void __fastcall TE7_DDraw::Flip()
{
  RECT R;
  FTimerOdd = !FTimerOdd;

  if (FActive == false)
   return;

  if (!FExclusive)
  {
    try
    {
      DrawBitmaps();
      GetWindowRect(FHandle, &R);
      DDTest(FPrimarySurface->Blt(&R, FBackSurface, &ScreenRect, 0, NULL), "Flip");
    }
    catch(Exception &E)
    {
      ErrorEvent("Flipping" + E.Message);
    };
  }
  else
  {
    try
    {
      if (FActive) {
        DrawBitmaps();
        do {
           Application->ProcessMessages();
        } while(!MakeItSo(FPrimarySurface->Flip(NULL, DDFLIP_WAIT)));
      }
    }
    catch(...)
    {
      ErrorEvent("Flipping");
    }
  }
}
//---------------------------------------------------------------------------
//TE7_DDSurface---------------------------------------------------------------
__fastcall TE7_DDSurface::TE7_DDSurface(TComponent *Owner)
 :TComponent(Owner)
{
  ShowBitmap = true;
}
//---------------------------------------------------------------------------
__fastcall TE7_DDSurface::~TE7_DDSurface()
{
  if (FDLLFile != NULL) FreeLibrary(FDLLFile);
}
//---------------------------------------------------------------------------
LPDIRECTDRAWSURFACE7 TE7_DDSurface::CreateBitmap(HANDLE Bitmap)
{
  DDSURFACEDESC2 SurfaceDesc;
  BITMAP BM;
  LPDIRECTDRAWSURFACE7 pdds;

  if (Bitmap == 0) throw Exception("CreateBitmap: No Bitmap");

  try
  {
    GetObject(Bitmap, sizeof(BM), &BM);

    memset(&SurfaceDesc, 0, sizeof(SurfaceDesc));
    SurfaceDesc.dwSize = sizeof(SurfaceDesc);
    SurfaceDesc.dwFlags = DDSD_CAPS | DDSD_HEIGHT | DDSD_WIDTH;
    SurfaceDesc.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN;
    SurfaceDesc.dwWidth = BM.bmWidth;
    SurfaceDesc.dwHeight = BM.bmHeight;

    if (AE_DDraw->ADirectDraw->CreateSurface(&SurfaceDesc, &pdds, NULL) != DD_OK)
      throw Exception("CreateBitmap: CreateSurface failed");

    DDCopyBitmap(pdds, Bitmap, Rect(0, 0, 0, 0));
  }
  catch(Exception &E)
  {
    AE_DDraw->ErrorEvent("CreateBitmap: " + E.Message);
  }

  return pdds;
}
//---------------------------------------------------------------------------
bool __fastcall TE7_DDSurface::CreateSurface(LPDIRECTDRAWSURFACE7 &pdds,
  AnsiString BitmapName, bool UsePalette)
{
  bool Result;
  DDCOLORKEY ColorKey;
  HRESULT hr;
  HANDLE Dib;

  if (UsePalette == true) {
    FPalette = DDLoadPalette(FDLLFile, AE_DDraw->ADirectDraw, BitmapName.c_str());
    AE_DDraw->PrimarySurface->SetPalette(FPalette);
  }

  Dib = Get_Dib(FDLLFile, BitmapName);
  pdds = CreateBitmap(Dib);

  if (!pdds) throw EDDError("CreateSurface: Failed");

  if (AE_DDraw->Exclusive == true) {
    DDSetColorKey(pdds, FTransparentColor);
    hr = pdds->SetColorKey(DDCKEY_SRCBLT, &ColorKey);
    if (hr != DD_OK)
      throw EDDError("CreateSurface: %d %s",
        OPENARRAY(TVarRec, (int(hr), GetOleError(hr))));
    else
      Result = true;
  }
  else {
    hr = DDSetColorKey(pdds, FTransparentColor);
    if (hr != DD_OK)
      throw EDDError("CreateSurface: %d %s",
        OPENARRAY(TVarRec, (int(hr), GetOleError(hr))));
    else
      Result = true;
  }

  return Result;
}
//---------------------------------------------------------------------------
void __fastcall TE7_DDSurface::DestroyObjects()
{
  DDDestroySurface(FDDSurface);
}
//---------------------------------------------------------------------------
void _fastcall TE7_DDSurface::DrawScene()
{
  AnsiString ErrStr = "DrawScene: %d \r %s \r %s";
  HRESULT hr;
  HDC DC;
  HBRUSH OldBrush, Brush;

  if (FBlankScene == true) {
    AE_DDraw->BackSurface->GetDC(&DC);
    Brush = CreateSolidBrush(FBackColor);
    OldBrush = SelectObject(DC, Brush);
    Rectangle (DC, 0, 0, AE_DDraw->ScreenRect.right, AE_DDraw->ScreenRect.bottom);
    SelectObject(DC, OldBrush);
    DeleteObject(Brush);
    AE_DDraw->BackSurface->ReleaseDC(DC);
  }
  if (FShowBitmap == true) {
    hr = AE_DDraw->BackSurface->BltFast(0, 0, FDDSurface, &FBackRect,
     DDBLTFAST_WAIT | DDBLTFAST_SRCCOLORKEY);

    if (!Windows::Succeeded(hr)) throw EDDError(ErrStr, OPENARRAY(TVarRec,
     (int(hr), GetOleError(hr), "DrawScene: Bitmap incorect size?")));
  }
  if (FOnDrawScene) {
    FOnDrawScene(this, AE_DDraw->BackSurface);
  }
}
//---------------------------------------------------------------------------
HANDLE __fastcall TE7_DDSurface::Get_Dib(HANDLE Instance, AnsiString S)
{
  HANDLE Result;
  UINT Flags;

  if (Instance != 0)
    Flags = LR_CREATEDIBSECTION;
  else
    Flags = LR_LOADFROMFILE | LR_CREATEDIBSECTION;

  Result = LoadImage((HANDLE)Instance, S.c_str(), IMAGE_BITMAP,
   0, 0, Flags);
  if (Result == 0)
    AE_DDraw->ErrorEvent("GetDib: Could not load bitmap" + S);

  return Result;
}
//---------------------------------------------------------------------------
HRESULT TE7_DDSurface::RestoreSurfaces(void)
{
  HRESULT Result;
  Result = FDDSurface->Restore();
  if (Result == DD_OK) DDReloadBitmap(FDDSurface, FDLLFile, FBackgroundBitmap);

  return(Result);
}
//---------------------------------------------------------------------------
void __fastcall TE7_DDSurface::SetupSurfaces(System::TObject *Sender)
{
  AE_DDraw = dynamic_cast <TE7_DDraw *>(Sender);

  if (FDLLName.Length() > 0) {
    if (!FDLLFile) FDLLFile = LoadLibrary(FDLLName.c_str());
   	if ((int) (FDLLFile) < 32) ShowMessage("Can't Load DLL:" + FDLLName);
    FDLLName = "";
    FDLLFile = NULL;
  }

  if(!CreateSurface(FDDSurface, FBackgroundBitmap, true))
   	throw EDDError("SetupSurfaces: No Surface: " +
     FBackgroundBitmap);
  else FBackRect = AE_DDraw->ScreenRect;

  if (FOnSetupSurfaces) FOnSetupSurfaces(this);
}
//---------------------------------------------------------------------------
void __fastcall TE7_DDSurface::WriteXY(int X, int Y, AnsiString S,
 COLORREF TextColor, COLORREF BackColor)
{
  HDC DC;
  FDDSurface->GetDC(&DC);
  SetBkColor(DC, BackColor);
  SetTextColor(DC, TextColor);
  TextOut(DC, X, Y, S.c_str(), S.Length());
  FDDSurface->ReleaseDC(DC);
}
//---------------------------------------------------------------------------
//TE7_Sprite-----------------------------------------------------------------
__fastcall TE7_Sprite::TE7_Sprite()
{
  Images = new TList();
  FPosition = 0;
  FSingle = true;
}
//---------------------------------------------------------------------------
__fastcall TE7_Sprite::~TE7_Sprite()
{
  delete Images;
}
//---------------------------------------------------------------------------
void __fastcall TE7_Sprite::AddImage(HANDLE Dll, AnsiString FileName)
{
  LPDIRECTDRAWSURFACE7 Temp;
  Temp = DDLoadBitmap(Dll, FileName, false, 0L);

  Images->Add(Temp);
  if (Images->Count > 1) FSingle = false;
}
//---------------------------------------------------------------------------
void __fastcall TE7_Sprite::AddImage(HANDLE Dll, AnsiString FileName,
 COLORREF Transparent)
{
  LPDIRECTDRAWSURFACE7 Temp;
  Temp = DDLoadBitmap(Dll, FileName, true, Transparent);

  Images->Add(Temp);
  if (Images->Count > 1) FSingle = false;
}
//---------------------------------------------------------------------------
void __fastcall TE7_Sprite::BlitFast(LPDIRECTDRAWSURFACE7 Dest,
 IPoint SourcePos, RECT SourceSize, LPRECT DestSize)
{
  if (!FSingle) {
    FPosition++;
    if (FPosition >= Images->Count) FPosition = 0;
  }
  DDBlitFast(Dest, (LPDIRECTDRAWSURFACE7)Images->Items[FPosition], SourcePos, SourceSize, DestSize);
}
//---------------------------------------------------------------------------
void __fastcall TE7_Sprite::BlitFx(LPDIRECTDRAWSURFACE7 Dest,
 IPoint SourcePos, RECT Size, DWORD Blt_Flags, LPDDBLTFX ddfx)
{
  if (!FSingle) {
    FPosition++;
    if (FPosition >= Images->Count) FPosition = 0;
  }
  DDBlitFx(Dest, (LPDIRECTDRAWSURFACE7)Images->Items[FPosition], SourcePos, Size, Blt_Flags, ddfx);
}
//---------------------------------------------------------------------------
//Register-------------------------------------------------------------------
namespace Engine7
{
  void __fastcall PACKAGE Register()
  {
    TComponentClass classes[2] = {__classid(TE7_DDraw), __classid(TE7_DDSurface)};
    RegisterComponents("Engine7", classes, 1);
  }
}
//---------------------------------------------------------------------------
